#include "pflash.h"
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef FREE_BSD
#include "getopt.h"
#else
#include <getopt.h>
#endif

#include <readline/readline.h>
#include <signal.h>

#ifdef FREE_BSD
#define DEFAULT_PORT "/dev/ttyd0"
#else
#define DEFAULT_PORT "/dev/ttyS0"
#endif

#define DEFAULT_SPEED "38400"
#define BUF_SIZE 1024
 


static int fd;

static unsigned char buffer[2048];

static unsigned char convertbuf[256];

static char char_version[4];
static char char_version_f[4];

static int buf_size = BUF_SIZE;



static int marker = 0xa5;
static int x80 = 0x80;

static int fmarker = 0;

static int size;

static int model,hardware,version,aversion,num_lit;

static int info_device = 1;

char *read_file = "prot1.zlx"; 

static int read_enable = 0;
char *write_file = "flash.zlx"; 

static int write_enable = 0;

static int speed = 38400;

static int keys = 1;

char *port_file = DEFAULT_PORT;

#ifndef FREE_BSD
static int option_index = 0;


static struct option long_options[] =
{
  {"help", 0, NULL, 'h'},               
  {"help", 0, NULL, '?'},               
  {"info", 0, NULL, 'i'},               
  {"test", 0, NULL, 't'},               
  {"read", 1, NULL, 'r'},		
  {"write", 1, NULL, 'w'},		
  {"port", 1, NULL, 'p'},                  
  {0, 0, 0, 0}
};
#endif




int calc_timeout(int byte_count) {

	return (8+2)*2*(byte_count+10)*1000000/speed;
}



int put_command(int fd, unsigned char *buffer, int buflen, int size_con);

int count_buf(unsigned char *buffer, int buflen)  {
    int fasa = 1,i=0,j=0;
    
    for (; i < buflen; i++) {
	if ((fasa == 1) && (buffer[i] == marker)) { fasa = 2; j++;}
	else if ((fasa == 2) && (buffer[i] == marker)) { fasa = 1; fmarker ++; }
	else { j++; fasa = 1;}
    }
    return j;
}



static void
pflash_sighandler(int sig)
{
    char buffer[4];
    int buflen = 0;
    int err = 10;
  

    buffer[0]=0x1; buflen=1;
    while (err) {
    
	if((buflen=put_command(fd, buffer, buflen, 4))==4) {printf(" \n"); close(fd); printf(" Device is switched to normal mode. \n"); exit(0);  }
	err --;
    }    
    printf(" \n");
    close(fd);
    exit(1);
}


static void
init_sighandler(void)
{
  struct sigaction act;

  act.sa_handler = pflash_sighandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  sigaction(SIGHUP, &act, 0);
  sigaction(SIGINT, &act, 0);
  sigaction(SIGCHLD, &act, 0);
  sigaction(SIGQUIT, &act, 0);
  sigaction(SIGTERM, &act, 0);
}





int calc_crc( int crc, int data) {
    int dat=0, af=0, j=1; 
							
    
    dat = data; 					
    
    for (; j <= 8; j++ ) {
	dat = dat << 1; 				
     
	af = crc; 					
    
	crc = crc >> 1; 				
    
	if ((dat & 0x100) != 0 ) { af = af ^ 1 ; 	
     
				    }
	if ((af & 1) != 0) { crc = crc ^ 0x8408;
							
     
				    }
	
    
    }
    return crc;
}


int to_prog (char *buffer){
    int hi_ccc=0,lo_ccc=0,i=0,ccc=0;

    for (; i < buf_size; i++ ) {
	ccc = calc_crc(ccc,i); 				
    
	lo_ccc = (ccc & 0xff);
	hi_ccc = ((ccc >> 8) & 0xff);			
    
	lo_ccc = (lo_ccc | 0x80);
	if ( lo_ccc == marker) { lo_ccc = x80;
 }
	buffer[i] = lo_ccc;		
	
    }
    return 1;
}


int 
translate_speed(int spd)
{
  int speed_c = 0;

  switch(spd)
  {
   case 9600:
    speed_c = B9600;
    break;
   case 19200:
    speed_c = B19200;
    break;
   case 38400:
    speed_c = B38400;
    break;
   case 57600:
    speed_c = B57600;
    break;
   case 115200:
    speed_c = B115200;
    break;
   case 230400:
    speed_c = B230400;
    break;
#ifndef FREE_BSD
   case 460800:
    speed_c = B460800;
    break;
#endif    
   default:
    printf("Bad baudrate %d.\n", spd);
    break;
  }
  return speed_c;
}


void
fd_setup(int fd, int speed, int flow)
{
  struct termios t;

  if (fd < 0)
  {
    perror("fd_setup");
    exit(1);
  }
	
  if (tcgetattr(fd, &t) < 0)
  {  

    printf(" There is an error of getting %s characteristics.\n",port_file);
    exit(1);
  }

  cfmakeraw(&t);
	
#ifndef FREE_BSD
  t.c_cflag &= ~CBAUD|~PARENB|~CSTOPB;
#else
  t.c_cflag &= ~PARENB|~CSTOPB;
#endif
  
#ifndef FREE_BSD
  t.c_cflag |= translate_speed(speed) | CS8 | CLOCAL;
#else
  t.c_cflag |= CS8 | CLOCAL; 
  cfsetspeed(&t,translate_speed(speed));
#endif
  t.c_oflag = 0; 
  t.c_lflag = 0; 
  
#ifndef FREE_BSD
  if (flow)
    t.c_cflag |= CRTSCTS;
  else
    t.c_cflag &= ~CRTSCTS;
#endif
  
  if (tcsetattr(fd, TCSANOW, &t) < 0)
  {

    printf(" There is an error of setting %s characteristics. (8bit data, 1 stop bit, parity none, speed 38.4K) \n",port_file);
    exit(1);
  }
  return;
}


int init_port(void) {
  int fd;

  if ((fd = open(port_file, O_RDWR | O_NDELAY | O_NOCTTY)) < 0 )
  { 

    printf(" The port %s could not be open. \n",port_file);
    exit(1);
  }
  
  tcflush(fd, TCIOFLUSH);

  return fd;
}


int make_frame(unsigned char *buf, int buflen) {
    unsigned char conv_buf[2048];
    int ccc, hi_ccc, lo_ccc, byte, i=0,j=0;

    ccc = 0;
    conv_buf[0] = 0xa5;
    for (; i < buflen; i++) {			
    
	conv_buf[++j]=buf[i];
	byte = buf[i];
    	ccc = calc_crc(ccc,byte); 		
    
	if (byte == marker) { conv_buf[++j]=byte; }
    }
    hi_ccc = ((ccc >> 8) & 0xff);
    lo_ccc = (ccc & 0xff);			
    
    if ( lo_ccc == marker ) {conv_buf[++j]=lo_ccc; conv_buf[++j]=lo_ccc;}
    else { conv_buf[++j]=lo_ccc; }
    if ( hi_ccc == marker ) {conv_buf[++j]=hi_ccc; conv_buf[++j]=hi_ccc;}
    else { conv_buf[++j]=hi_ccc; }

    for (i=0; i <= j; i++) {
	buf[i] = conv_buf[i];
    }						
    
    return ++j;
}

int m_read(int fd, char *databuf, int size_con){

  fd_set rfd;
  
  struct timeval tv;


  
  int cur_len=0, retval,len = 0, count_len = 0;

  int sc = size_con;
  
  size_con = size_con * 2;

  while (1)
  {  
    FD_ZERO(&rfd); 
    FD_SET(fd, &rfd);
    tv.tv_sec = 0;

    tv.tv_usec = calc_timeout(2);
    
    retval=select(fd+1, &rfd, NULL, NULL, &tv);    
    
    
    if (retval)
    {
	if(FD_ISSET(fd, &rfd)) {len = read(fd, databuf, size_con);}
	
	count_len += count_buf(databuf, len);

	size_con -= len;
	cur_len  += len;
	
	databuf  += len;


        if (( count_len >= sc)  )
	{


	    if ( fmarker == 0) {
        	return cur_len;
	    } else {
		fmarker = 0;
        	return cur_len;
	    }
	}
    } else 
    
    {
	fmarker = 0;
        if (cur_len < 4) { 

	    return -1; }
	else {

	    return cur_len;
	}
    }
  }

}


int m_read1(int fd, unsigned char *databuf, unsigned char *readbuf, int size_con)
{

  fd_set rfd;
  struct timeval tv;
  
  int cur_len=0, retval,len = 0, count_len = 0, str_len = 0, fasa = 0,j=0;
  unsigned char *start = databuf;

int cut_marker_buf(unsigned char *buffer, unsigned char *readbuf, int buflen)  {
    int i=0;

    for (; i < buflen; i++) {
	if ((readbuf[i] == marker) && (fasa == 0) ) { fasa = 1; buffer[j] = marker; j++;}
	else if ((fasa == 1) && (readbuf[i] == marker)) { fasa = 2; buffer[j] = marker; j++; }
	else if ((fasa == 2) && (readbuf[i] == marker)) { fasa = 1; }
	else if (fasa == 1) {  buffer[j] = readbuf[i]; j++;}
	else { }
    }
    buffer  += buflen;
    return j;
}


  size_con = 11;
  while (1)
  {  
    FD_ZERO(&rfd); 
    FD_SET(fd, &rfd);
    tv.tv_sec = 1;

    tv.tv_usec = calc_timeout(10+4);

    retval=select(fd+1, &rfd, NULL, NULL, &tv);    
    
    
    if (retval)
    {
	if(FD_ISSET(fd, &rfd)) {len = read(fd, readbuf, size_con);}
	
    
	count_len = cut_marker_buf(databuf, readbuf, len);
	cur_len  += len;
	readbuf += len;

        if (count_len >= 10)
	{
	    str_len = (start[10] & 0xff);

	    if(cur_len >= 11+str_len+4) { 
	    	fmarker = 0;


        	return (11+str_len+4);
	    }
	} 
    } else 
    {



        if (cur_len < 10) { 

	return -1;}
	
	else {

	    return cur_len;
	}
    }
  }

}


void get_info_device(unsigned char *buffer, int buflen) {
    unsigned char *name;

    name = buffer+11;

    model = (buffer[2]<<8)|(buffer[1] & 0xff);
    hardware = (buffer[4]<<8)|(buffer[3] & 0xff);
    version = (buffer[6]<<8)|(buffer[5] & 0xff);
    num_lit = (buffer[10] & 0xff);
    
    if ( (buflen - 10 - num_lit) >= 5 ) {
    aversion = (buffer[12+num_lit]<<8)|(buffer[11+num_lit] & 0xff);}
    size = (buffer[9]<<16)|(buffer[8]<<8)|(buffer[7] & 0xff);
    buffer[11+num_lit] = '\0';
    
    snprintf(char_version,4,"%d",version);
    if (version == 65535) {printf(" There is a CRC error in program %s\n",name); if(read_enable) {exit(1);}}
    else {
  if(info_device) printf ("\n%s version V%c.%s  \n",name,char_version[0],&char_version[1]);
    }

}


int test_frame(unsigned char *buffer, int buflen) {
    int ccc=0,hi_ccc,lo_ccc,new_hi_ccc,new_lo_ccc,i=1;
							    
    
     if ( marker != buffer[0] ) { return -1;}
    hi_ccc = buffer[buflen-1]; lo_ccc = buffer[buflen-2];	
    
    for (; i < buflen-2; i++) {			
    
	    ccc = calc_crc(ccc,buffer[i]); 		
    
    }
	new_hi_ccc = ((ccc >> 8) & 0xff);
	new_lo_ccc = (ccc & 0xff);			
    
    if ( (new_lo_ccc != lo_ccc) || (new_hi_ccc != hi_ccc) ) { return 0;}
    return 1;
}




int convert_buf(unsigned char *buffer, int buflen)  {
    unsigned char conv_buf[2048];
    int fasa = 0,i=0,j=0;
    
    for (; i < buflen; i++) {
	if ((buffer[i] == marker) && (fasa == 0)) { fasa = 1; conv_buf[j++] = buffer[i]; }
	else if ((fasa == 1) && (buffer[i] == marker)) { fasa = 2; conv_buf[j++] = buffer[i];}
	else if ((fasa == 2) && (buffer[i] == marker)) { fasa = 1; }
	else { conv_buf[j++] = buffer[i]; }
    }

    for (i=0; i < j; i++) {
	buffer[i] = conv_buf[i];
    }						
    
    return j;
}




int put_command(int fd, unsigned char *buffer, int buflen, int size_con) {
    unsigned char conv_buf[2048];
    int error=1,error_count=0,i=0,conv_len=buflen,n=0;		
    fd_set wfd,rfd;
    struct timeval tv;
    
    for (; i < buflen; i++) {conv_buf[i] = buffer[i];}		
    
	conv_len = make_frame(conv_buf,conv_len);			
    
    while (error)
    {
    FD_ZERO(&wfd); 
    FD_ZERO(&rfd); 
    FD_SET(fd, &wfd);
    FD_SET(fd, &rfd);
    tv.tv_sec = 0;

    tv.tv_usec = calc_timeout(300);
	
	select(fd+1, NULL, &wfd, NULL, &tv);    
	if(FD_ISSET(fd, &wfd)) {n = write(fd, conv_buf, conv_len);



	}

	select(fd+1, &rfd, NULL, NULL, &tv);    
	
    	if ( n && (errno == EAGAIN) ) {
	
	sleep(1);
	} else if ( n != conv_len ) {
	    
	} else {
	    if(FD_ISSET(fd, &rfd)) {buflen=m_read(fd, buffer, size_con);}
    
    }
	if (buflen < size_con)
	{ 
	    
    
	    error_count++;
	} else {

	buflen=convert_buf(buffer, buflen);		
    
	if(test_frame(buffer, buflen)) { error = 0;} else {error_count++;printf("Frame error\n");}
	}
	if (error_count > 2) { 
     
				return -1;}
    }
    return buflen;
}



int put_command1(int fd, unsigned char *buffer, unsigned char *readbuf, int buflen, int size_con) {
    unsigned char conv_buf[2048];
    int error=1,error_count=0,i=0,conv_len=buflen,n,rv;		
    
    for (; i < buflen; i++) {conv_buf[i] = buffer[i];}		
    
	conv_len = make_frame(conv_buf,conv_len);			
    
    while (error)
    {

    rv = to_prog(buffer);



    if ((n = write(fd, buffer, buf_size)) != buf_size) {
	printf("Saving : tried to write %d bytes, wrote %d\n",buf_size, n);
    }


	sleep(1);    
	n = write(fd, conv_buf, conv_len);


    
    
	if ( n && (errno == EAGAIN) ) {
	
    
	sleep(1);
	} else if ( n != conv_len ) {
	
    
	} else {
	
	buflen=m_read1(fd, buffer, convertbuf, size_con);    
	
    
    }
	if (buflen < size_con)
	{ 
	    
    
	    error_count++;
	} else {

	buflen=convert_buf(buffer, buflen);		
    
	if(test_frame(buffer, buflen)) { error = 0;} else {error_count++;printf("Frame error\n");}
	}
	if (error_count > 6) { 
     
				return -1;}
    }
    return buflen;
}




#ifndef FREE_BSD
void usage(const char *name) {
    printf("\n");
    printf("Usage: %s [pflash-options]\n", name);
    printf("    [-h|--help|-?]         Print this help.\n");
    printf("    default: enable\n");
    printf("\n");
    printf("    [-i|--info|-t|--test]   Not print info about device and software.\n");
    printf("    default: disable\n");
    printf("\n");
    printf("    [-r|--read path/file]  Save flash in the file.\n");
    printf("\n");
    printf("    [-w|--write path/file] Write the flash from the file.\n");
    printf("\n");
    printf("    [-p|--port path/ttySn]   Set port of UART device.\n");
    exit(1);
}
#else
void usage(const char *name) {
    printf("\n");
    printf("Usage: %s [pflash-options]\n", name);
    printf("    [-h|-?]         Print this help.\n");
    printf("    default: disable\n");
    printf("\n");
    printf("    [-i|-t]   Not print info about device and firmware.\n");
    printf("    default: disable\n");
    printf("\n");
    printf("    [-r path/file]  Save flash in the file.\n");
    printf("\n");
    printf("    [-w path/file] Write the flash from the file.\n");
    printf("\n");
    printf("    [-p path/ttyXX]   Set port of UART device.\n");
    exit(1);
 }
#endif




int 
main(int argc, char **argv)
{ 
    int ffd, n, buflen=0,cbuf=0,rv,test_error=0,opt,cond=1,oldvalue;
    struct stat s;
    static unsigned char *p_buffer;
    int ccc=0,hi_ccc,lo_ccc,new_hi_ccc,new_lo_ccc,i,per_sent=0,offset1 = 0;
    char *line;

    char *name=argv[0];
    char *nametmp=name;

    while (*nametmp) { 
	if ( *(nametmp++) == '/' ) name=nametmp; 
    }

void restart(void) {
    buffer[0]=0x1; buflen=1;
    if((buflen=put_command(fd, buffer, buflen, 4))==4) {}
    else exit(1);
    close(fd);
}

void progress_bar(int value,char *word_pr) {
    int prog,value1; 
    char str[20];
    if (value > 100) value = 100;
    if (value != oldvalue) {
	value1 = (int) (value*20/100) + 1;
	for (prog=0;prog < value1;prog++) str[prog] = '.'; 
	for (;prog < 20;prog++) str[prog] = ' '; str[20] = '\0';
	fprintf(stderr,"%s%3d%c    %s",str,value,'%',word_pr);
    }
    oldvalue=value;
}
    
printf("pflash (v1.0) - Program of reading and writing the flash of ZELAX\n");

  
#define OPTIONS_STRING "hip:r:tw:X"

#ifndef FREE_BSD
  while ((opt = getopt_long(argc, argv, OPTIONS_STRING,
                            long_options, &option_index)) != -1)
#else
 while ((opt = getopt(argc, argv, OPTIONS_STRING)) != -1) 
#endif			    
			    
  {   
    switch(opt)
    {
     case 'h':
      usage(name);
      break;
      
     case 'i':
      info_device = 0;
      break;
      
     case 'p':
      keys = 0; 
      port_file = strdup(optarg);
      break;
      
     case 'r':
      read_file = strdup(optarg);
      read_enable = 1;
      break;

     case 't':
      info_device = 0;
      break;

     case 'w':
      write_file = strdup(optarg);
      write_enable = 1;
      break;

     case '?':
      usage(name);
      break;

    default:
      break;
    }
  }
    if ((opt == -1)&&(keys == 1)) { usage(name); }
    
    init_sighandler();
    
    fd = init_port();  
    
    fd_setup(fd, speed, 0);




    buffer[0]=0; buflen=1;
    if((buflen=put_command1(fd, buffer, convertbuf, buflen, 11)) > 11) {get_info_device(buffer, buflen);}
    else { printf ("No access to device! Check, is it connected to the COM-port or not.\n");
           printf ("The port in the device should work at the speed of 38400 bps.\n");
           printf ("Read the chapter on loading a program into the device in the device manual.\n");
    restart(); exit(1);}


if(read_enable) {
    write_enable = 0; 
    if ( (ffd = creat(read_file, (S_IREAD | S_IWRITE) & ~S_IRWXG )) < 0 ) { perror("open_file"); restart();  exit(1); }
    
    while (cbuf != size) {
        buflen=4;
	buffer[0]=2;
	buffer[1] = (cbuf & 0xff);
	buffer[2] = ((cbuf >> 8) & 0xff);
	buffer[3] = ((cbuf >> 16) & 0xff);
	if((buflen=put_command(fd, buffer, buflen, 131)) > 130 ) {

								    progress_bar((cbuf+128)*100/size,"Read\r"); 
	    if ((n = write(ffd, buffer+1, buflen-3)) != buflen-3) {
		printf("Saving flash into file: tried to write %d bytes, wrote %d\n",buf_size, n);}


	} else { printf(" Device access error. Reading is aborted.\n"); restart(); exit(1);}
	cbuf = cbuf+128;				
    
    }
    close(ffd);
    fprintf (stderr,"\n");
}

if(write_enable) {

    if ((ffd = open(write_file, O_RDONLY )) < 0 ) { perror("open_file"); restart();  exit(1); }
    rv = fstat(ffd, &s);

    if ( s.st_size == size ) { 
    }
    else { printf ("Size error of the file %s. \n",write_file); test_error=1; }
    p_buffer = malloc(s.st_size);
    
    buflen=read(ffd,p_buffer,s.st_size);

    hi_ccc = p_buffer[s.st_size-1]; lo_ccc = p_buffer[s.st_size-2];	
    
    for (i=0; i < s.st_size-2; i++) {			
    
	    ccc = calc_crc(ccc,p_buffer[i]); 		
    
    }
    new_hi_ccc = ((ccc >> 8) & 0xff);
    new_lo_ccc = (ccc & 0xff);				
    
    if ( (new_lo_ccc != lo_ccc) || (new_hi_ccc != hi_ccc) ) { printf("The CRC error in the file %s. \n",write_file); test_error=1; }
    else { 
    }

    if ( ((p_buffer[0]<<8)|(p_buffer[1] & 0xff)) == model ) { 
     }
    else { printf ("Firmware %s is not compatible with this device.\n",write_file); test_error=1;  }
    if ( ((p_buffer[2]<<8)|(p_buffer[3] & 0xff)) ==hardware ) { 
    }
    else { printf (" Firmware %s is not compatible with this device.\n",write_file); test_error=1; }
    if ( ((p_buffer[4]<<8)|(p_buffer[5] & 0xff)) >= aversion ) { 
    }
    else { printf(" The given device can run with firmware version more V%d. \nThe current version V%c.%s \n",aversion,char_version[0],&char_version[1]); test_error=1;}
    
    

    if ( !test_error) {


    snprintf(char_version_f,4,"%d",((p_buffer[4]<<8)|(p_buffer[5] & 0xff)));
    while (cond)
    {
      printf ("Do you want to write the file with version V%c.%s into the flash? [Y/N]\n",char_version_f[0],&char_version_f[1]);

      line = readline("> ");

      if (!strncmp(line, "Y", 1) || !strncmp(line, "y", 1))
      {
	cond = 0;
      }
      else { restart(); exit(0);}
    }

	buffer[0]=3; buffer[1]=0xa2; buffer[2]=0x49; buffer[3]=0xb6; buflen=4;
	if((buflen=put_command(fd, buffer, buflen, 4))==4) { if (buffer[1] == 0x55) {}}
	else {printf(" Error of flash erasing.\n\n"); restart(); exit(1);}

while (per_sent < 100 )
	{
	    buffer[0]=4; buflen=1; 
	    
	    if((buflen=put_command(fd, buffer, buflen, 4))<3) 
	{printf("Device access error. Erasing is aborted.\n");restart(); exit(1);}
	    if (buffer[1]) {	
				per_sent = buffer[2];
				if ( per_sent > 100 ) per_sent = 100;
				progress_bar(per_sent,"Erase\r"); 
			} else {printf(" Error of the erase operation.\n\n"); restart(); exit(1);}
	}
 
	fprintf(stderr,"                                      \r");
	offset1 = 0;
	
	while ( offset1 < size ) {
    	    buflen=132;
	    buffer[0]=5;
	    buffer[1] = (offset1 & 0xff);
	    buffer[2] = ((offset1 >> 8) & 0xff);
	    buffer[3] = ((offset1 >> 16) & 0xff);
	    for (i=4; i < 132; i++) {
		buffer[i]=p_buffer[offset1-4+i];
	    }
	    offset1 =  offset1 + 128;
	    if((buflen=put_command(fd, buffer, buflen, 4))==4) { 
    
			    progress_bar(offset1*100/size,"Write     \r");	    
								    
	    }
	    else {printf("Device access error. Writing is aborted. \n"); restart(); exit(1);}
	} 
    } else {printf(" Operation is aborted. \n"); restart(); exit(1);}
    free(p_buffer);
    close(ffd);
    fprintf (stderr,"\n");
}
    

    restart();
    return 0;
}
